#include "dependency.h"

#include <regex>

#include "project.h"

constexpr std::string_view downloads("downloads");
constexpr std::string_view sources("sources");
constexpr std::string_view build("build");

constexpr std::string_view acceptJson("accept: application/json");
const std::regex http("^https?:\\/\\/");
const std::regex repo("(^https?):\\/\\/([\\w\\d\\.\\-\\_]+)\\/([\\w\\d\\-\\_]+)\\/([\\w\\d\\-\\_]+)");

Dependency::Dependency(
    const std::string& path,
    Type type,
    const std::optional<std::string>& name,
    const std::optional<std::string>& version
):
    path(path),
    type(type),
    name(name),
    version(version),
    location()
{}

Dependency::Type Dependency::getType() const {
    return type;
}

std::optional<std::string> Dependency::getName() const {
    return name;
}

std::optional<std::string> Dependency::getVersion() const {
    return version;
}

std::optional<std::filesystem::path> Dependency::getLocation() const {
    if (location.empty())
        return std::nullopt;
    else
        return location;
}

bool Dependency::prepare(const std::filesystem::path& source, const std::filesystem::path& destination) {
    std::smatch results;
    if (std::regex_search(path, results, repo)) {
        Project::info(path + " appears to be a git repository");
        const std::string& protocol = results[1];
        const std::string& host = results[2];
        const std::string& owner = results[3];
        const std::string& repo = results[4];

        std::filesystem::path temp;
        bool success = downloadRepo(destination, protocol, host, owner, repo, temp);
        if (success) {
            Project::info("Successfully obtained project " + path);

            switch (type) {
                case Type::automatic:
                    if (std::filesystem::exists(temp/"mason.json"))
                        type = Type::mason;
                    else
                        type = Type::simple;

                    break;
                case Type::mason:
                    if (std::filesystem::exists(temp/"mason.json"))
                        break;
                    else {
                        Project::warn("Project " + path + " is supposed to me a mason project, but no mason.json was found in the project directory");
                        success = false;
                    }

                    break;
                case Type::simple:
                    break;
            }
        }

        return success;
    } else {
        Project::info("checking project at path " + path);
        std::filesystem::directory_entry srcDir(source / path);
        if (!srcDir.exists()) {
            Project::error("Project at " + path + " doesn't exist");
            return false;
        }
    }

    return true;
}

bool Dependency::downloadRepo(
    const std::filesystem::path& destination,
    const std::string& protocol,
    const std::string& host,
    const std::string& owner,
    const std::string& repo,
    std::filesystem::path& out
) {
    nlohmann::json info;
    Project::info("Trying Gitea v1 API");
    bool res = repoGiteaApi1(protocol, host, owner, repo, info);

    if (res) {
        nlohmann::json::const_iterator itr = info.find("default_branch");
        if (itr != info.end() && itr->is_string()) {
            std::string branchName = *itr;
            Project::info("Gitea v1 API seem to have worked");
            Project::info("Default branch is " + branchName);

            std::string fileName = branchName + ".tar.gz";
            std::string url = protocol + "://" + host + "/api/v1/repos/" + owner + "/" + repo + "/archive/" + fileName;
            std::filesystem::path archivePath = destination/downloads/fileName;
            res = download(url, archivePath);
            if (res) {
                Project::info("Successfully downloaded " + archivePath.string());

                location = destination/sources;
                res = extract(archivePath, location);
                if (res) {
                    out = destination/build/repo;
                    if (!std::filesystem::is_directory(out)) {
                        Project::error("Extracted archive " + fileName + " but the content is unexpected");
                        res = false;
                    }
                } else {
                    Project::error("Couldn't extract archive " + fileName);
                }

                if (!res)
                    location.clear();

                return res;
            }
        } else {
            res = false;
        }
    }

    if (!res) {
        Project::warn("Gitea v1 API didn't work");
    }

    return res;
}

bool Dependency::repoGiteaApi1(
    const std::string& protocol,
    const std::string& host,
    const std::string& owner,
    const std::string& repo,
    nlohmann::json& json
) {
    bool result = false;
    std::string url = protocol + "://" + host + "/api/v1/repos/" + owner + "/" + repo;
    std::string data;
    CURLcode code = httpGet(url, data, {acceptJson});
    if (code == CURLE_OK) {
        try {
            json = nlohmann::json::parse(data);
            result = true;
        } catch (const nlohmann::json::exception& e) {
            Project::warn(e.what());
        }
    }

    return result;
}

CURLcode Dependency::httpGet(const std::string& url, std::string& result, const std::list<std::string_view>& headers) {
    CURL* curl = curl_easy_init();
    curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
    curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);
    curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1L);

    struct curl_slist* curlHeaders = nullptr;
    for (const std::string_view& header : headers)
        curlHeaders = curl_slist_append(curlHeaders, header.data());

    curl_easy_setopt(curl, CURLOPT_HTTPHEADER, curlHeaders);
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, writeString);
    curl_easy_setopt(curl, CURLOPT_WRITEDATA, &result);

    curl_easy_setopt(curl, CURLOPT_VERBOSE, 1L);
    curl_easy_setopt(curl, CURLOPT_DEBUGFUNCTION, trace);
    curl_easy_setopt(curl, CURLOPT_DEBUGDATA, this);

    CURLcode res = curl_easy_perform(curl);

    curl_easy_cleanup(curl);
    return res;
}


bool Dependency::download(const std::string& url, const std::filesystem::path& destination) {
    std::filesystem::create_directories(destination.parent_path());
    if (std::filesystem::exists(destination)) {
        Project::minor("File " + destination.string() + " already exists, will be overwritten");

        if (std::filesystem::is_directory(destination))
            std::filesystem::remove_all(destination);
    }

    bool result = false;
    CURL* curl = curl_easy_init();
    curl_easy_setopt(curl, CURLOPT_URL, url.c_str());
    curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1L);
    curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, writeFile);

    curl_easy_setopt(curl, CURLOPT_VERBOSE, 1L);
    curl_easy_setopt(curl, CURLOPT_DEBUGFUNCTION, trace);
    curl_easy_setopt(curl, CURLOPT_DEBUGDATA, this);

    FILE* pagefile = fopen(destination.c_str(), "wb");
    if (pagefile) {
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, pagefile);
        CURLcode res = curl_easy_perform(curl);

        fclose(pagefile);
        if (res != CURLE_OK) {
            Project::warn("Couldn't download file " + url + ": " + curl_easy_strerror(res));
        } else {
            uint32_t code;
            curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &code);
            if (code == 200)
                result = true;
            else
                Project::warn("Couldn't download file " + url + ": response code " + std::to_string(code));
        }
    } else {
        Project::error(std::string("Couldn't open output file ") + destination.c_str());
    }

    curl_easy_cleanup(curl);

    if (!result) {
        Project::minor("Removing " + destination.string() + " since the donwload failed");
        std::filesystem::remove_all(destination);
    }

    return result;
}

size_t Dependency::writeString(void* data, size_t size, size_t nmemb, void* pointer) {
    size_t finalSize = size * nmemb;
    std::string* string = static_cast<std::string*>(pointer);
    string->append(static_cast<char*>(data), finalSize);

    return finalSize;
}

size_t Dependency::writeFile(void* data, size_t size, size_t nmemb, void* file) {
    size_t written = fwrite(data, size, nmemb, (FILE *)file);
    return written;
}

int Dependency::trace(CURL* handle, curl_infotype type, char* data, size_t size, void* clientp) {
    switch (type) {
        case CURLINFO_TEXT: {
            std::string message(data, size);
            if (message[size - 1] == '\n')
                message = message.substr(0, size - 1);

            Project::debug(message);
        } break;
        default:
            break;
    }

    return 0;
}

bool Dependency::extract(const std::filesystem::path& source, const std::filesystem::path& destination) const {
    int flags = ARCHIVE_EXTRACT_TIME;
    flags |= ARCHIVE_EXTRACT_PERM;
    flags |= ARCHIVE_EXTRACT_ACL;
    flags |= ARCHIVE_EXTRACT_FFLAGS;

    struct archive* a = archive_read_new();
    struct archive* ext = archive_write_disk_new();
    struct archive_entry *entry;
    archive_read_support_format_all(a);
    archive_read_support_filter_all(a);
    archive_write_disk_set_options(ext, flags);
    archive_write_disk_set_standard_lookup(ext);

    bool result = true;
    bool readOpen = false;
    bool writeOpen = false;
    int r = archive_read_open_filename(a, source.c_str(), 10240);
    if (r) {
        Project::major("Couldn't open file " + source.string());
        result = false;
    } else {
        readOpen = true;
    }

    while (result) {
        r = archive_read_next_header(a, &entry);
        if (r == ARCHIVE_EOF)
            break;

        if (r < ARCHIVE_OK)
            Project::major(archive_error_string(a));

        if (r < ARCHIVE_WARN)
            break;

        std::string fileName(archive_entry_pathname(entry));
        std::filesystem::path filePath = destination/fileName;
        Project::debug("Extracting " + filePath.string());
        if (std::filesystem::exists(filePath))
            Project::minor(filePath.string() + " exists, overwriting");

        archive_entry_set_pathname_utf8(entry, filePath.c_str());

        r = archive_write_header(ext, entry);
        if (r < ARCHIVE_OK) {
            Project::major(archive_error_string(ext));
        } else if (archive_entry_size(entry) > 0) {
            writeOpen = true;
            r = copy(a, ext);
            if (r < ARCHIVE_OK)
                Project::major(archive_error_string(ext));

            if (r < ARCHIVE_WARN)
                break;
        }
        r = archive_write_finish_entry(ext);
        if (r < ARCHIVE_OK)
            Project::major(archive_error_string(ext));

        if (r < ARCHIVE_WARN)
            break;
    }

    if (readOpen)
        archive_read_close(a);

    if (writeOpen)
        archive_write_close(ext);

    archive_read_free(a);
    archive_write_free(ext);

    return result;
}

int Dependency::copy(struct archive* ar, struct archive* aw) const {
    int r;
    const void *buff;
    size_t size;
    la_int64_t offset;

    while (true) {
        r = archive_read_data_block(ar, &buff, &size, &offset);
        if (r == ARCHIVE_EOF)
            return ARCHIVE_OK;

        if (r < ARCHIVE_OK)
            return r;

        r = archive_write_data_block(aw, buff, size, offset);
        if (r < ARCHIVE_OK) {
            Project::major(archive_error_string(aw));
            return r;
        }
    }
}


